
unit DiscUtil;

interface

uses Windows, SysUtils, Controls, ShellAPI, Classes, Forms;

//--- Error handling -----------------------------------------------------------

function DiscErrorMessage (Drive : Char) : string;
function FormatVolumeSerialNumber (N : integer) : string;
function Retry (Drive : Char) : boolean;
function SysErrorMessage (ErrorCode: Integer): string;
function SysErrorMessageParams (ErrorCode: Integer;
                const Params : array of string): string;


//--- Disc volume information --------------------------------------------------

type
  TVolumeInformation = record
    VolumeName  : string;
    VolumeSerialNumber,
    MaximumComponentLength,
    {$IFDEF VER120}
    FileSystemFlags : dword;
    {$ELSE}
    FileSystemFlags : dword;
    {$ENDIF}
    FileSystemName : string;
  end;

function GetNetworkVolumeName (D : char): string;
function GetVolumeInformation (D : char; var V : TVolumeInformation) : boolean;
function GetVolumeName (D : char) : string;

//--- System image lists -------------------------------------------------------


var
  SmallImages,
  LargeImages : TImageList;

//--- Types of disc drive ------------------------------------------------------

type
  TDriveType = (dtUnknown, dtNoDrive, dtFloppy, dtFixed, dtNetwork, dtCDROM,
                dtRAM, dtFloppy3, dtFloppy5);

const
  DriveNames : array [TDriveType] of string[9] =
    ('Unknown', 'None', 'Floppy', 'Fixed', 'Network', 'CD-ROM',
     'RAM', '3 Floppy', '5 Floppy');

function FloppyDriveSize (Drive : char) : TDriveType;

function GetDriveType (Drive : Char) : TDriveType;

type
  TDriveShellInfo = record
    Icon : hIcon;
    Image : integer;
    DisplayName,
    TypeName : string
  end;

procedure GetDriveShellInfo (Drive : Char; var Info : TDriveShellInfo);

//--- Information on available drives ------------------------------------------

procedure GetLogicalDriveList (List : TStrings);

type
  TDiscFreeSpace = record
    SectorsPerCluster,
    BytesPerSector,
    NumberOfFreeClusters,
    TotalNumberOfClusters,
    TotalSpace,
    FreeSpace : {$IFDEF VER120} DWORD {$ELSE}integer {$ENDIF}
  end;

function GetDiscFreeSpace (Drive : char; var D : TDiscFreeSpace) : boolean;

//------------------------------------------------------------------------------

implementation

//--- Internal

procedure RealizeLength(var S: string);
begin
  SetLength (S, StrLen (PChar(S)))
end;

//--- Error handling

function SysErrorMessageParams (ErrorCode: Integer; const Params : array of string): string;

const
  L = 255;
begin
  SetLength (Result, L);
  FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_ARGUMENT_ARRAY,
      nil, ErrorCode, 0, PChar(Result), L, @Params);
  RealizeLength (Result)
end;

function SysErrorMessage (ErrorCode: Integer): string;
begin
  Result := SysErrorMessageParams (ErrorCode, [''])
end;

function FormatVolumeSerialNumber (N : integer) : string;
begin
  Result := Format ('%X-%X', [longrec(N).hi, longrec(N).lo])
end;


function DiscErrorMessage (Drive : Char) : string;
begin
  Result := Format ('%s:\ is not accessible.'#13#10#13#10+'%s',
                     [uppercase(Drive), SysErrorMessage (GetLastError)])
end;

function Retry (Drive : Char) : boolean;
begin
  Result := Application.MessageBox (
              PChar (DiscErrorMessage (Drive)),
              PChar (Application.Title),
              mb_RetryCancel or mb_IconError) = idRetry
end;

//--- Volume Information


function GetVolumeInformation (D : char; var V : TVolumeInformation) : boolean;
var
  O : integer;
begin
  O := SetErrorMode (SEM_FAILCRITICALERRORS);
  try
    with V do
    begin
      SetLength (VolumeName, MAX_PATH);
      SetLength (FileSystemName, MAX_PATH);
      VolumeSerialNumber := 0;
      MaximumComponentLength := 0;
      FileSystemFlags := 0;
      Result := Windows.GetVolumeInformation (PChar (D+':\'), PChar (VolumeName), MAX_PATH,
        @VolumeSerialNumber, MaximumComponentLength, FileSystemFlags,
        PChar (FileSystemName), MAX_PATH);
      RealizeLength (VolumeName);
      RealizeLength (FileSystemName)
    end
  finally
    SetErrorMode (O)
  end
end;

function GetVolumeName (D : char) : string;
var
  T : TVolumeInformation;
begin
  if GetVolumeInformation (D, T) then
    Result := T.VolumeName
  else
    Result := ''
end;

function GetNetworkVolumeName (D : char): string;
var
  {$IFDEF VER120}
  L : dword ;
  {$ELSE}
  L : Integer ;
  {$ENDIF}
begin
  L := MAX_PATH;
  SetLength (Result, L);
  if WNetGetConnection (PChar(D+':'#0), PChar(Result), L) = NO_ERROR then
    RealizeLength (Result)
  else
    Result := GetVolumeName(D)
end;

//--- floppy disc size determination

function FloppyDriveSize (Drive : char) : TDriveType;
type
  PDIOC_REG = ^TDIOC_Registers;
  TDIOC_Registers = record
    Reg_EBX, Reg_EDX, Reg_ECX, Reg_EAX, Reg_EDI, Reg_ESI, Reg_Flags : DWORD
  end;
const
  VWIN32_DIOC_DOS_INT13 = 4;  // Performs Interrupt 13h commands.
var

  H : THandle;
  R : TDIOC_Registers;
  C : DWORD;
begin
  Result := dtFloppy;
  H := CreateFile ('\\.\VWIN32', 0, 0, nil, 0, 0, 0);
  if H <> INVALID_HANDLE_VALUE then
  try
    R.Reg_EAX := $800;                    // service 8 in AH
    R.Reg_EDX := ord (Upcase(Drive)) - Ord('A');  // drive number in DL
    if DeviceIOControl (H, VWIN32_DIOC_DOS_INT13, @R, SizeOf (R), @R, SizeOf (R), C, nil)
       and (R.Reg_Flags and 1 = 0) then   // clear CF indicates success

      if R.Reg_EBX and $FF < 3 then       // drive type in BL
         Result := dtFloppy5              // 1 = 360K, 2 = 1.2MB
      else
         Result := dtFloppy3              // 3 = 720K  4 = 1.44MB  5 = 2.88MB
  finally
    CloseHandle (H)
  end
end;

//--- Available drive information

function GetDiscFreeSpace (Drive : char; var D : TDiscFreeSpace) : boolean;
var
  O : integer;
begin
  FillChar (D, Sizeof (TDiscFreeSpace), 0);
  O := SetErrorMode (SEM_FAILCRITICALERRORS);

  try
    with D do
    begin
      Result := Windows.GetDiskFreeSpace (PChar (Drive + ':\'), SectorsPerCluster,
        BytesPerSector, NumberOfFreeClusters, TotalNumberOfClusters);
      FreeSpace := BytesPerSector*SectorsPerCluster*NumberOfFreeClusters;
      TotalSpace := BytesPerSector*SectorsPerCluster*TotalNumberOfClusters
    end
  finally
    SetErrorMode (O)
  end
end;

procedure GetLogicalDriveList (List : TStrings);
var
  Num  : integer;
  Bits : set of 0..25;

begin
  List.Clear;
  integer (Bits) := Windows.GetLogicalDrives;
  for Num := 0 to 25 do
    if Num in Bits then
      List.Add (Char (Num + Ord('A')) + ':\')
end;

function GetDriveType (Drive : Char) : TDriveType;
begin
  Result := TDriveType (Windows.GetDriveType(PChar(Drive + ':\')))
end;

procedure GetDriveShellInfo (Drive : Char; var Info : TDriveShellInfo);
var
  SHFileInfo : TSHFileInfo;
begin
  ShGetFileInfo (PChar (Drive + ':\'), 0, SHFileInfo, SizeOf (TSHFileInfo),

     SHGFI_TYPENAME or SHGFI_DISPLAYNAME or SHGFI_SYSICONINDEX or SHGFI_ICON);
  with Info do
  begin
    Icon  := SHFileInfo.hIcon;
    Image := SHFileInfo.iIcon;
    DisplayName := SHFileInfo.szDisplayName;
    TypeName := SHFileInfo.szTypeName
  end
end;

//--- unit start-up and close-down

function GetImageList (Option : integer) : TImageList;
var
  SHFileInfo : TSHFileInfo;
begin
  Result := TImageList.Create (nil);
  with Result do
  begin
    ShareImages := true;

    Handle := ShGetFileInfo ('*.*', 0, SHFileInfo, SizeOf (TSHFileInfo),
                   Option or SHGFI_SYSICONINDEX)
  end
end;

initialization
  LargeImages := GetImageList (SHGFI_LARGEICON);
  SmallImages := GetImageList (SHGFI_SMALLICON)
finalization
  LargeImages.Free;
  SmallImages.Free
end.


